from wxPython.wx import *

from Thuban.UI.resource import GetBitmapResource
from Thuban.UI.graph import Graph

class Sliders(Graph):
    """A set of graphical markers representing values that can be handled.

    Handling works like this:

    - Clicking on an empty spot creates a new, unselected,  marker at that spot.

    - Clicking on a currently existing marker toggles its selection status

    - Dragging a selected marker moves ALL currently selected markers. If they
    fall outside the permitted arnge, they are deleted.

    - Double-clicking any marker, selected or unselected, removes it.

    Values passed to the class at creation and values returned for marker
    positions are normalized to [0,1].
    """
    # Everthing internal to Sliders is handled in screen coordinates.
    # Input is normalized to [0,1] as with everything derived from Graph
    # Output, idem.
    def __init__(self, parent, pos = None, width = 400, height = 10):
        Graph.__init__(self, parent, width, height)
        """Create a set of markers.

        'pos' are initial positions (in [0,1]) for the markers. Use GetList()
        to query the marker positions. Use GetSelectedIndices() to query the
        list of currently selected marker indices corresponding to GetList.
        """
        self.marker_bitmap = GetBitmapResource(
            "triangle_marker_green", wxBITMAP_TYPE_XPM)
        self.marker_dc = wxMemoryDC()
        self.marker_dc.SelectObject(self.marker_bitmap)

        self.marker_bitmap_selected = GetBitmapResource(
            "triangle_marker_red", wxBITMAP_TYPE_XPM)
        self.marker_dc_selected = wxMemoryDC()
        self.marker_dc_selected.SelectObject(self.marker_bitmap_selected)

        self.x = None
        self.index = None
        self.selected = []

        self.offset = self.border * self.w
        self.scalefactor = self.w / (1.0 + 2.0 * self.border)

        if len(pos) > 0:
            self.pos = [ p * self.scalefactor + self.offset for p in pos ]
        else:
            self.pos = [ pos * self.scalefactor + self.offset ]

        EVT_LEFT_DOWN( self, self.OnLeftDown )
        EVT_LEFT_UP( self, self.OnLeftUp )
        EVT_MOTION( self, self.OnMotion )
        EVT_LEAVE_WINDOW( self, self.OnLeaveWindow )
        EVT_LEFT_DCLICK( self, self.OnLeftDClick)

    def GetList(self):
        """Returns the values pointed at by the markers.

        These are always in [0,1].
        """
        l = [ (p - self.offset) / self.scalefactor for p in self.pos ]
        l.sort()
        return l

    def GetSelectedIndices(self):
        """Returns the indices of selected markers.

        Selected positions are thus:

        [ GetList()[i] for i in GetSelectedIndices() ]
        """
        # construct select indicator, zip it to pos, sort the tuples
        # and extract the new indices.
        # TODO: better ideas?
        st = [0]*len( self.pos )
        for si in self.selected:
            st[si] = 1
        pt = zip( self.pos, st )
        pt.sort()
        return [ i for i in range(len(pt)) if pt[i][1] ]

    def __findSlider(self, xy):
        """
        Return (found,i) where found indicates whether a marker i is at xy.
        """
        i = 0
        found = False
        for p in self.pos:
            if abs(xy.x - p) < xy.y / 2.0 + 1: #a "triangular" search
                found = True
                break
            i += 1
        return (found, i)

    def getNum(self):
        """
        Returns the current number of markers.
        """
        return len(self.pos)

    def OnLeftDClick(self,event):
        """
        Delete the marker under the cursor.
        """
        xy = event.GetPosition()
        (found, i) = self.__findSlider( xy )
        if found:
            self.__remove_markers( [i] )
            self.refresh()

    def OnLeftDown(self,event):
        """
        Add or remove a marker from the selected list.

        Afterwards, all selected markers can be moved by dragging.
        """
        xy = event.GetPosition()
        (found, i) = self.__findSlider( xy )
        if found: # highlight it
            self.draw_marker_selected( self.pos[i] )
            self.index = i
        else:
            self.pos.append( xy.x )
        self.x = xy.x
        self.found = found
        self.refresh()
        
    def OnLeftUp(self, event):
        """
        Set a new marker, or move the previously selected ones.

        If markers fall outside the range, they are deleted.
        """
        evx = event.GetPosition().x
        moved = abs( evx - self.x ) >= 2
        if self.found:
            if not moved: # just select/deselect it
                if self.index in self.selected:
                    self.selected.remove(self.index)
                    self.draw_marker(self.pos[self.index])
                else:
                    self.selected.append(self.index)
                    self.draw_marker_selected(self.pos[self.index])
                self.found = False
            elif len(self.selected) > 0 and self.index in self.selected:
                # move selected only when one of them is being moved
                to_remove = []
                for i in self.selected[:]:
                    # Move all selected
                    newpos = self.pos[i] + (evx - self.x)
                    # Detect outsiders
                    if not self.offset <= newpos <= self.w - self.offset:
                        to_remove.append(i)
                    else:
                        self.pos[i] = newpos
                # Remove outsiders after movement
                self.__remove_markers( to_remove )
        self.refresh()

    def __remove_markers( self, to_remove ):
        """
        Remove markers from current position and selection lists.
        """
        to_remove.sort()
        to_remove.reverse()
        self.pos = [ self.pos[i]
                     for i in range(len(self.pos))
                     if i not in to_remove ] # better way to pop many idxs?
        s = self.selected[:]
        s.sort()
        for r in to_remove:
            news = []
            for si in s:
                if si == r:
                    pass
                elif si > r:
                    news.append( si - 1 )
                else:
                    news.append( si )
            s = news
        self.selected = s

    def OnMotion(self, event):
        # TODO: How do you drag a bitmap over another in wxWindows???
        pass

    def OnLeaveWindow(self, event):
        # TODO: Should this be a "deselect all"?
        pass

    def draw_marker(self, x, drawfunc = wxCOPY):
        """
        Stamps a marker at x.
        """
        dc = wxPaintDC(self)
        dc.Blit( int(x - 4), 0,
                 self.marker_bitmap.GetWidth(),
                 self.marker_bitmap.GetHeight(),
                 self.marker_dc, 0, 0, drawfunc)

    def draw_marker_selected(self, x, drawfunc = wxCOPY):
        """
        Stamps a "selected"-style marker at x.
        """
        dc = wxPaintDC(self)
        dc.Blit( int(x - 4), 0,
                 self.marker_bitmap_selected.GetWidth(),
                 self.marker_bitmap_selected.GetHeight(),
                 self.marker_dc_selected, 0, 0, drawfunc)

    def __repr__(self):
        return "Slider Set:\n\tnormalized positions : "+ repr([
            (p - self.offset) / self.scalefactor for p in self.pos
            ]) + "\n" +\
            "\tselected indices : " + repr(self.selected) + "\n"

    def refresh(self, dc = None):
        """
        Repaint the marker set distinguishing selected and unselected.
        """
        if dc is None:
            dc = wxPaintDC(self)
            dc.Clear()
        for p in self.pos:
            self.draw_marker( p )
        for i in self.selected:
            self.draw_marker_selected( self.pos[i] )


